class Field:
    def __init__(self,s,lok,toks):
        self.toks = toks
        self.level = toks["level"]
        self.title = toks["title"]
        self.list_of_values = {}
        self.length = None
        self.precision = None
        self.signed = False
        self.compressed = False
        self.signed_trailing = False

        if toks.get("type_numeric") is not None:
            type_descriptor = toks.get("type_numeric")
            self.length = int(type_descriptor.get("length","0"))
            if type_descriptor.get("precision"):
                self.datatype = "decimal"
                self.precision = int(type_descriptor.get("precision"))
            else:
                self.datatype = "int"
            
            # signed
            if toks.get("signed"):
                self.signed = True
                if toks.get("trailing"):
                    self.signed_trailing = True

            # compressed
            if toks.get("comp3"):
                self.compressed = True

            # validate
            if self.length is None and self.precision is None:
                raise ParseException(f"can't parse PIC for {toks}")
        elif toks.get("type_string"):
            # a string PIC
            type_descriptor = toks.get("type_string")
            self.length = int(type_descriptor.get("length","0"))

            self.datatype = "str"

            # check if we have a list of possible values. convert that to a nice constants array
            if toks.get("values"):
                for val in toks.get("values"):
                    self.list_of_values[val.get("title")] = val.get("value")
        else:
            # no type descriptor
            raise ParseException(f"unknown data type: {toks}")

    def __repr__(self):
        return_value = f"title: {self.title}\n- type: {self.datatype}\n- length: {self.length}\n- precision {self.precision}"
        if self.list_of_values != {}:
            for k,v in self.list_of_values.items():
                return_value += f"\n  - {k}:{v}"
        return return_value




# class representing a field group
class FieldGroup:
    def __init__(self,s,lok,toks):
        self.level = toks["level"]
        self.title = toks["title"]
        self.redefine_target = toks.get("redefine_target")
        self.repeat = toks.get("occurs",1)
        self.children = []

    def __repr__(self):
        return_value = f"""
title: {self.title}
- type: group
- redefines: {self.redefine_target}
"""
        return return_value




import copy
from .parser import stmt
from .field_group import FieldGroup
def parse_file(filename:str) -> FieldGroup:
    result:FieldGroup = stmt.parseFile(filename,parseAll=True)
    # rearrange in a tree structure
    return _list_to_tree(result)

def parse_string(contents:str) -> FieldGroup:
    result:FieldGroup = stmt.parseString(contents,parseAll=True)
    # rearrange in a tree structure
    return _list_to_tree(result)

# utility function to rearrange the fields in a tree structure based on the field IDs
def _relocate_field(root,field_list):
    root.children = []
    field_list.pop(0)
    while len(field_list)>0:
        field = field_list[0]
        # print(f"{field} under root: {root}")
        if field.level > root.level:
            root.children.append(_relocate_field(copy.copy(field),field_list))
        else:
            break
    return root

def _list_to_tree(field_list):
    return _relocate_field(copy.copy(field_list[0]),field_list.copy())

def _pprint_tree(root:FieldGroup,level=0):
    indent = "".join(['  ']*level)
    print(f"{indent}{root.title}")
    if type(root)==FieldGroup:
        for child in root.children:
            _pprint_tree(child,level+1)

